from importlib.util import find_spec

from daml._internal.metrics.ber import BER
from daml._internal.metrics.divergence import Divergence
from daml._internal.metrics.uap import UAP

__all__ = ["BER", "Divergence", "UAP"]

if (
    find_spec("tensorflow") is not None
    and find_spec("tensorflow_probability") is not None
):  # pragma: no cover
    from . import outlier_detection

    __all__ += ["outlier_detection"]

if find_spec("torch") is not None:  # pragma: no cover
    from daml._internal.metrics.sufficiency import Sufficiency

    __all__ += ["Sufficiency"]

del find_spec



import logging
from typing import Dict, Tuple

import numpy as np
import tensorflow as tf

from daml._alibi_detect.base import (
    BaseDetector,
    FitMixin,
    ThresholdMixin,
    outlier_prediction_dict,
)
from daml._alibi_detect.models.tensorflow.autoencoder import AE
from daml._alibi_detect.models.tensorflow.trainer import trainer
from daml._alibi_detect.utils._types import OptimizerTF
from daml._alibi_detect.utils.tensorflow.prediction import predict_batch

logger = logging.getLogger(__name__)


class OutlierAE(BaseDetector, FitMixin, ThresholdMixin):
    def __init__(
        self,
        threshold: float = None,
        ae: tf.keras.Model = None,
        encoder_net: tf.keras.Model = None,
        decoder_net: tf.keras.Model = None,
        data_type: str = None,
    ) -> None:
        """
        AE-based outlier detector.

        Parameters
        ----------
        threshold
            Threshold used for outlier score to determine outliers.
        ae
            A trained tf.keras model if available.
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class if no 'ae' is specified.
        decoder_net
            Layers for the decoder wrapped in a tf.keras.Sequential class if no 'ae' is specified.
        data_type
            Optionally specify the data type (tabular, image or time-series). Added to metadata.
        """
        super().__init__()

        if threshold is None:
            logger.warning(
                "No threshold level set. Need to infer threshold using `infer_threshold`."
            )

        self.threshold = threshold

        # check if model can be loaded, otherwise initialize AE model
        if isinstance(ae, tf.keras.Model):
            self.ae = ae
        elif isinstance(encoder_net, tf.keras.Sequential) and isinstance(
            decoder_net, tf.keras.Sequential
        ):
            self.ae = AE(encoder_net, decoder_net)
        else:
            raise TypeError(
                "No valid format detected for `ae` (tf.keras.Model) "
                "or `encoder_net`, `decoder_net` (tf.keras.Sequential)."
            )

        # set metadata
        self.meta["detector_type"] = "outlier"
        self.meta["data_type"] = data_type
        self.meta["online"] = False

    def fit(
        self,
        X: np.ndarray,
        loss_fn: tf.keras.losses = tf.keras.losses.MeanSquaredError(),
        optimizer: OptimizerTF = tf.keras.optimizers.Adam,
        epochs: int = 20,
        batch_size: int = 64,
        verbose: bool = True,
        log_metric: Tuple[str, "tf.keras.metrics"] = None,
        callbacks: tf.keras.callbacks = None,
    ) -> None:
        """
        Train AE model.

        Parameters
        ----------
        X
            Training batch.
        loss_fn
            Loss function used for training.
        optimizer
            Optimizer used for training.
        epochs
            Number of training epochs.
        batch_size
            Batch size used for training.
        verbose
            Whether to print training progress.
        log_metric
            Additional metrics whose progress will be displayed if verbose equals True.
        callbacks
            Callbacks used during training.
        """
        # train arguments
        args = [self.ae, loss_fn, X]
        optimizer = optimizer() if isinstance(optimizer, type) else optimizer
        kwargs = {
            "optimizer": optimizer,
            "epochs": epochs,
            "batch_size": batch_size,
            "verbose": verbose,
            "log_metric": log_metric,
            "callbacks": callbacks,
        }

        # train
        trainer(*args, **kwargs)

    def infer_threshold(
        self,
        X: np.ndarray,
        outlier_type: str = "instance",
        outlier_perc: float = 100.0,
        threshold_perc: float = 95.0,
        batch_size: int = int(1e10),
    ) -> None:
        """
        Update threshold by a value inferred from the percentage of instances considered to be
        outliers in a sample of the dataset.

        Parameters
        ----------
        X
            Batch of instances.
        outlier_type
            Predict outliers at the 'feature' or 'instance' level.
        outlier_perc
            Percentage of sorted feature level outlier scores used to predict instance level outlier.
        threshold_perc
            Percentage of X considered to be normal based on the outlier score.
        batch_size
            Batch size used when making predictions with the autoencoder.
        """
        # compute outlier scores
        fscore, iscore = self.score(X, outlier_perc=outlier_perc, batch_size=batch_size)
        if outlier_type == "feature":
            outlier_score = fscore
        elif outlier_type == "instance":
            outlier_score = iscore
        else:
            raise ValueError(
                "`outlier_score` needs to be either `feature` or `instance`."
            )

        # update threshold
        self.threshold = np.percentile(outlier_score, threshold_perc)

    def feature_score(self, X_orig: np.ndarray, X_recon: np.ndarray) -> np.ndarray:
        """
        Compute feature level outlier scores.

        Parameters
        ----------
        X_orig
            Batch of original instances.
        X_recon
            Batch of reconstructed instances.

        Returns
        -------
        Feature level outlier scores.
        """
        fscore = np.power(X_orig - X_recon, 2)
        return fscore

    def instance_score(
        self, fscore: np.ndarray, outlier_perc: float = 100.0
    ) -> np.ndarray:
        """
        Compute instance level outlier scores.

        Parameters
        ----------
        fscore
            Feature level outlier scores.
        outlier_perc
            Percentage of sorted feature level outlier scores used to predict instance level outlier.

        Returns
        -------
        Instance level outlier scores.
        """
        fscore_flat = fscore.reshape(fscore.shape[0], -1).copy()
        n_score_features = int(np.ceil(0.01 * outlier_perc * fscore_flat.shape[1]))
        sorted_fscore = np.sort(fscore_flat, axis=1)
        sorted_fscore_perc = sorted_fscore[:, -n_score_features:]
        iscore = np.mean(sorted_fscore_perc, axis=1)
        return iscore

    def score(
        self, X: np.ndarray, outlier_perc: float = 100.0, batch_size: int = int(1e10)
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute feature and instance level outlier scores.

        Parameters
        ----------
        X
            Batch of instances.
        outlier_perc
            Percentage of sorted feature level outlier scores used to predict instance level outlier.
        batch_size
            Batch size used when making predictions with the autoencoder.

        Returns
        -------
        Feature and instance level outlier scores.
        """
        # reconstruct instances
        X_recon = predict_batch(X, self.ae, batch_size=batch_size)

        # compute feature and instance level scores
        fscore = self.feature_score(X, X_recon)  # type: ignore[arg-type]
        iscore = self.instance_score(fscore, outlier_perc=outlier_perc)

        return fscore, iscore

    def predict(
        self,
        X: np.ndarray,
        outlier_type: str = "instance",
        outlier_perc: float = 100.0,
        batch_size: int = int(1e10),
        return_feature_score: bool = True,
        return_instance_score: bool = True,
    ) -> Dict[Dict[str, str], Dict[np.ndarray, np.ndarray]]:
        """
        Predict whether instances are outliers or not.

        Parameters
        ----------
        X
            Batch of instances.
        outlier_type
            Predict outliers at the 'feature' or 'instance' level.
        outlier_perc
            Percentage of sorted feature level outlier scores used to predict instance level outlier.
        batch_size
            Batch size used when making predictions with the autoencoder.
        return_feature_score
            Whether to return feature level outlier scores.
        return_instance_score
            Whether to return instance level outlier scores.

        Returns
        -------
        Dictionary containing ``'meta'`` and ``'data'`` dictionaries.
            - ``'meta'`` has the model's metadata.
            - ``'data'`` contains the outlier predictions and both feature and instance level outlier scores.
        """
        # compute outlier scores
        fscore, iscore = self.score(X, outlier_perc=outlier_perc, batch_size=batch_size)
        if outlier_type == "feature":
            outlier_score = fscore
        elif outlier_type == "instance":
            outlier_score = iscore
        else:
            raise ValueError(
                "`outlier_score` needs to be either `feature` or `instance`."
            )

        # values above threshold are outliers
        outlier_pred = (outlier_score > self.threshold).astype(int)

        # populate output dict
        od = outlier_prediction_dict()
        od["meta"] = self.meta
        od["data"]["is_outlier"] = outlier_pred
        if return_feature_score:
            od["data"]["feature_score"] = fscore
        if return_instance_score:
            od["data"]["instance_score"] = iscore
        return od



from abc import ABC, abstractmethod
from typing import Callable, Dict, Generic, List, TypeVar

TOutput = TypeVar("TOutput", bound=dict)
TMethods = TypeVar("TMethods")
TCallable = TypeVar("TCallable", bound=Callable)


class EvaluateMixin(ABC, Generic[TOutput]):
    @abstractmethod
    def evaluate(self) -> TOutput:
        """Abstract method to calculate metric based off of constructor parameters"""


class MethodsMixin(ABC, Generic[TMethods, TCallable]):
    """
    Use this mixin to define a mapping of functions to method names which
    can be queried by the user and called internally with the appropriate
    method name as the key.

    Explicitly defining the Callable generic helps with type safety and
    hinting for function signatures and recommended but optional.

    e.g.:

    def _mult(x: float, y: float) -> float:
        return x * y

    class MyMetric(MethodsMixin[Callable[float, float], float]):

        def _methods(cls) -> Dict[str, Callable[float, float], float]:
            return {
                "ADD": lambda x, y: x + y,
                "MULT":  _mult,
                ...
            }

    Then during evaluate, you can call the method specified with the getter.

    e.g.:

        def evaluate(self):
            return self._method(x, y)

    The resulting class can be used like so.

    m = MyMetric(1.0, 2.0, "ADD")
    m.evaluate()       #  returns 3.0
    m.method           #  returns "ADD"
    MyMetric.methods() #  returns "['ADD', 'MULT']
    m.method = "MULT"
    m.evaluate()       #  returns 2.0
    """

    @classmethod
    @abstractmethod
    def _methods(cls) -> Dict[str, TCallable]:
        """Abstract method returning available method functions for class"""

    @property
    def _method(self) -> TCallable:
        return self._methods()[self.method]

    @classmethod
    def methods(cls) -> List[str]:
        return list(cls._methods().keys())

    @property
    def method(self) -> str:
        return self._method_key

    @method.setter
    def method(self, value: TMethods):
        self._set_method(value)

    def _set_method(self, value: TMethods):
        """This setter is to fix pyright incorrect detection of
        incorrectly overriding the 'method' property"""
        if value not in self.methods():
            raise KeyError(
                f"Specified method not available for class ({self.methods()})."
            )
        self._method_key = value



from enum import Enum
from typing import Dict, Iterable, List, Optional

from .missing_optional_dependency import ERROR_TYPES


class Framework(str, Enum):
    PYTORCH = "pytorch"
    TENSORFLOW = "tensorflow"
    KEOPS = "keops"
    SKLEARN = "sklearn"


try:
    import tensorflow as tf  # noqa
    import tensorflow_probability as tfp  # noqa

    has_tensorflow = True
except ImportError:
    has_tensorflow = False

try:
    import torch  # noqa

    has_pytorch = True
except ImportError:
    has_pytorch = False

try:
    import pykeops  # noqa
    import torch  # noqa

    has_keops = True
except ImportError:
    has_keops = False

# Map from backend name to boolean value indicating its presence
HAS_BACKEND = {
    "tensorflow": has_tensorflow,
    "pytorch": has_pytorch,
    "sklearn": True,
    "keops": has_keops,
}


def _iter_to_str(iterable: Iterable[str]) -> str:
    """Correctly format iterable of items to comma separated sentence string."""
    items = [f"`{option}`" for option in iterable]
    last_item_str = f"{items[-1]}" if not items[:-1] else f" and {items[-1]}"
    return ", ".join(items[:-1]) + last_item_str


class BackendValidator:
    def __init__(
        self, backend_options: Dict[Optional[str], List[str]], construct_name: str
    ):
        """Checks for required sets of backend options.

        Takes a dictionary of backends plus extra dependencies and generates correct error messages if they are unmet.

        Parameters
        ----------
        backend_options
            Dictionary from backend to list of dependencies that must be satisfied. The keys are the available options
            for the user and the values should be a list of dependencies that are checked via the `HAS_BACKEND` map
            defined in this module. An example of `backend_options` would be `{'tensorflow': ['tensorflow'], 'pytorch':
            ['pytorch'], None: []}`.This would mean `'tensorflow'`, `'pytorch'` or `None` are available backend options.
            If the user passes a different backend they will receive and error listing the correct backends. In
            addition, if one of the dependencies in the `backend_option` values is missing for the specified backend
            the validator will issue an error message telling the user what dependency bucket to install.
        construct_name
            Name of the object that has a set of backends we need to verify.
        """
        self.backend_options = backend_options
        self.construct_name = construct_name

    def verify_backend(self, backend: str):
        """Verifies backend choice.

        Verifies backend is implemented and that the correct dependencies are installed for the requested backend. If
        the backend is not implemented or a dependency is missing then an error is issued.

        Parameters
        ----------
        backend
            Choice of backend the user wishes to initialize the alibi-detect construct with. Must be one of the keys
            in the `self.backend_options` dictionary.

        Raises
        ------
        NotImplementedError
            If backend is not a member of `self.backend_options.keys()` a `NotImplementedError` is raised. Note `None`
            is a valid choice of backend if it is set as a key on `self.backend_options.keys()`. If a backend is not
            implemented for an alibi-detect object then it should not have a key on `self.backend_options`.
        ImportError
            If one of the dependencies in `self.backend_options[backend]` is missing then an ImportError will be thrown
            including a message informing the user how to install.
        """
        if backend not in self.backend_options:
            self._raise_implementation_error(backend)

        dependencies = self.backend_options[backend]
        missing_deps = []
        for dependency in dependencies:
            if not HAS_BACKEND[dependency]:
                missing_deps.append(dependency)

        if missing_deps:
            self._raise_import_error(missing_deps, backend)

    def _raise_import_error(self, missing_deps: List[str], backend: str):
        """Raises import error if backend choice has missing dependency."""

        optional_dependencies = list(
            ERROR_TYPES[missing_dep] for missing_dep in missing_deps
        )
        optional_dependencies.sort()
        missing_deps_str = _iter_to_str(missing_deps)
        error_msg = (
            f"{missing_deps_str} not installed. Cannot initialize and run {self.construct_name} "
            f"with {backend} backend."
        )
        pip_msg = (
            ""
            if not optional_dependencies
            else (
                f"The necessary missing dependencies can be installed using "
                f'`pip install alibi-detect[{" ".join(optional_dependencies)}]`.'
            )
        )
        raise ImportError(f"{error_msg} {pip_msg}")

    def _raise_implementation_error(self, backend: str):
        """Raises NotImplementedError error if backend choice is not implemented."""

        backend_list = _iter_to_str(self.backend_options.keys())
        raise NotImplementedError(
            f"{backend} backend not implemented. Use one of {backend_list} instead."
        )



from typing import Any, Literal, Tuple

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree as mst
from scipy.spatial.distance import pdist, squareform
from sklearn.neighbors import NearestNeighbors

EPSILON = 1e-5


def minimum_spanning_tree(X: np.ndarray) -> Any:
    """
    Returns the minimum spanning tree from a NumPy image array.

    Parameters
    ----------
    X: np.ndarray
        Numpy image array

    Returns
    -------
        Data representing the minimum spanning tree
    """
    # All features belong on second dimension
    X = X.reshape((X.shape[0], -1))
    # We add a small constant to the distance matrix to ensure scipy interprets
    # the input graph as fully-connected.
    dense_eudist = squareform(pdist(X)) + EPSILON
    eudist_csr = csr_matrix(dense_eudist)
    return mst(eudist_csr)


def get_classes_counts(labels: np.ndarray) -> Tuple[int, int]:
    """
    Returns the classes and counts of from an array of labels

    Parameters
    ----------
    label: np.ndarray
        Numpy labels array

    Returns
    -------
        Classes and counts

    Raises
    ------
    ValueError
        If the number of unique classes is less than 2
    """
    classes, counts = np.unique(labels, return_counts=True)
    M = len(classes)
    if M < 2:
        raise ValueError("Label vector contains less than 2 classes!")
    N = np.sum(counts).astype(int)
    return M, N


def compute_neighbors(
    A: np.ndarray,
    B: np.ndarray,
    k: int = 1,
    algorithm: Literal["auto", "ball_tree", "kd_tree"] = "auto",
) -> np.ndarray:
    """
    For each sample in A, compute the nearest neighbor in B

    Parameters
    ----------
    A, B : np.ndarray
        The n_samples and n_features respectively
    k : int
        The number of neighbors to find
    algorithm : Literal
        Tree method for nearest neighbor (auto, ball_tree or kd_tree)

    Note
    ----
        Do not use kd_tree if n_features > 20

    Returns
    -------
    List:
        Closest points to each point in A and B

    See Also
    --------
    :func:`sklearn.neighbors.NearestNeighbors`
    """

    nbrs = NearestNeighbors(n_neighbors=k + 1, algorithm=algorithm).fit(B)
    nns = nbrs.kneighbors(A)[1]
    nns = nns[:, 1]

    return nns



"""
This module contains the implementation of Dp Divergence
using the Fast Nearest Neighbor and Minimum Spanning Tree algorithms
"""

from typing import Any, Callable, Dict, Literal

import numpy as np

from daml._internal.metrics.base import EvaluateMixin, MethodsMixin

from .utils import compute_neighbors, minimum_spanning_tree


def _mst(data: np.ndarray, labels: np.ndarray) -> int:
    mst = minimum_spanning_tree(data).toarray()
    edgelist = np.transpose(np.nonzero(mst))
    errors = np.sum(labels[edgelist[:, 0]] != labels[edgelist[:, 1]])
    return errors


def _fnn(data: np.ndarray, labels: np.ndarray) -> int:
    nn_indices = compute_neighbors(data, data)
    errors = np.sum(np.abs(labels[nn_indices] - labels))
    return errors


_METHODS = Literal["FNN", "MST"]
_FUNCTION = Callable[[np.ndarray, np.ndarray], int]


class Divergence(EvaluateMixin, MethodsMixin[_METHODS, _FUNCTION]):
    """
    Calculates the estimated divergence between two datasets

    Parameters
    ----------
    data_a : np.ndarray
        Array of images or image embeddings to compare
    data_b : np.ndarray
        Array of images or image embeddings to compare

    See Also
    --------
        For more information about this divergence, its formal definition,
        and its associated estimators see https://arxiv.org/abs/1412.6534.

    Warning
    -------
        MST is very slow in this implementation, this is unlike matlab where
        they have comparable speeds
        Overall, MST takes ~25x LONGER!!
        Source of slowdown:
        conversion to and from CSR format adds ~10% of the time diff between
        1nn and scipy mst function the remaining 90%
    """

    def __init__(
        self,
        data_a: np.ndarray,
        data_b: np.ndarray,
        method: _METHODS = "FNN",
    ) -> None:
        self.data_a = data_a
        self.data_b = data_b
        self._set_method(method)

    @classmethod
    def _methods(cls) -> Dict[str, _FUNCTION]:
        return {"FNN": _fnn, "MST": _mst}

    def evaluate(self) -> Dict[str, Any]:
        """
        Calculates the divergence and any errors between the datasets

        Returns
        -------
        Dict[str, Any]
            dp : float
                divergence value between 0.0 and 1.0
            errors : int
                the number of differing edges
        """
        N = self.data_a.shape[0]
        M = self.data_b.shape[0]

        stacked_data = np.vstack((self.data_a, self.data_b))
        labels = np.vstack([np.zeros([N, 1]), np.ones([M, 1])])

        errors = self._method(stacked_data, labels)
        dp = max(0.0, 1 - ((M + N) / (2 * M * N)) * errors)
        return {"dpdivergence": dp, "error": errors}



"""Functionality for optional importing
This module provides a way to import optional dependencies. In the case that the user imports some functionality from
alibi-detect that is not usable due to missing optional dependencies this code is used to allow the import but replace
it with an object that throws an error on use. This way we avoid errors at import time that prevent the user using
functionality independent of the missing dependency.
"""

from importlib import import_module
from string import Template
from typing import Any, List, Optional, Union

err_msg_template = Template(
    (
        "Attempted to use $object_name without the correct optional dependencies installed. To install "
        + "the correct optional dependencies, run `pip install alibi-detect[$missing_dependency]` "
        + "from the command line. For more information, check the Installation documentation "
        + "at https://docs.seldon.io/projects/alibi-detect/en/stable/overview/getting_started.html."
    )
)


"""Mapping used to ensure correct pip install message is generated if a missing optional dependency is detected. This
dict is used to control two behaviours:

1. When we import objects from missing dependencies we check that any `ModuleNotFoundError` or `ImportError`
    corresponds to a missing optional dependency by checking the name of the missing dependency is in `ERROR_TYPES`. We
    then map this name to the corresponding optional dependency bucket that will resolve the issue.
2. Some optional dependencies have multiple names such as `torch` and `pytorch`, instead of enforcing a single
    naming convention across the whole code base we instead use `ERROR_TYPES` to capture both cases. This is done right
    before the pip install message is issued as this is the most robust place to capture these differences.
"""
ERROR_TYPES = {
    "prophet": "prophet",
    "tensorflow_probability": "tensorflow",
    "tensorflow": "tensorflow",
    "torch": "torch",
    "pytorch": "torch",
    "keops": "keops",
    "pykeops": "keops",
}


class MissingDependency:
    """Missing Dependency Class.

    Used to replace any object that requires unmet optional dependencies. Attribute access or calling the __call__
    method on this object will raise an error.
    """

    def __init__(
        self,
        object_name: str,
        err: Union[ModuleNotFoundError, ImportError],
        missing_dependency: str = "all",
    ):
        """Metaclass for MissingDependency classes.

        Parameters
        ----------
        object_name
            Name of object we are replacing
        err
            Error to be raised when the class is initialized or used
        missing_dependency
            Name of missing dependency required for object
        """
        self.missing_dependency = missing_dependency
        self.object_name = object_name
        self.err = err

    @property
    def err_msg(self) -> str:
        """Generate error message informing user to install missing dependencies."""
        return err_msg_template.substitute(
            object_name=self.object_name, missing_dependency=self.missing_dependency
        )

    def __getattr__(self, key):
        """Raise an error when attributes are accessed."""
        raise ImportError(self.err_msg) from self.err

    def __call__(self, *args, **kwargs):
        """If called, raise an error."""
        raise ImportError(self.err_msg) from self.err


def import_optional(module_name: str, names: Optional[List[str]] = None) -> Any:
    """Import a module that depends on optional dependencies
    Note: This function is used to import modules that depend on optional dependencies. Because it mirrors the python
    import functionality its return type has to be `Any`. Using objects imported with this function can lead to
    misspecification of types as `Any` when the developer intended to be more restrictive.

    Parameters
    ----------
    module_name
        The module to import
    names
        The names to import from the module. If None, all names are imported.

    Returns
    -------
    The module or named objects within the modules if names is not None. If the import fails due to a \
    `ModuleNotFoundError` or `ImportError` then the requested module or named objects are replaced with instances of \
    the MissingDependency class above.
    """
    try:
        module = import_module(module_name)
        if names is not None:
            objs = tuple(getattr(module, name) for name in names)
            return objs if len(objs) > 1 else objs[0]
        return module
    except (ImportError, ModuleNotFoundError) as err:
        if err.name is None:
            raise err
        dep_name, *_ = err.name.split(".")
        if str(dep_name) not in ERROR_TYPES:
            raise err
        missing_dependency = ERROR_TYPES[dep_name]
        if names is not None:
            missing_dependencies = tuple(
                MissingDependency(
                    missing_dependency=missing_dependency, object_name=name, err=err
                )
                for name in names
            )
            return (
                missing_dependencies
                if len(missing_dependencies) > 1
                else missing_dependencies[0]
            )
        return MissingDependency(
            missing_dependency=missing_dependency, object_name=module_name, err=err
        )



import math
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, NamedTuple, Optional, Tuple

import keras
import numpy as np

from daml import _alibi_detect
from daml._internal.models.tensorflow.alibi import create_model


class OutlierType(str, Enum):
    """
    Enum to determine outliers by instance or by feature
    """

    INSTANCE = "instance"
    FEATURE = "feature"

    def __str__(self) -> str:
        return str.__str__(self)  # pragma: no cover


class ThresholdType(Enum):
    """
    Enum of threshold types for outlier detection
    """

    VALUE = "value"
    PERCENTAGE = "percentage"


class Threshold(NamedTuple):
    """
    NamedTuple to specify the threshold value and type for outlier detection

    Parameters
    ----------
    value : float
        The threshold to determine an outlier
    type : ThresholdType
        Whether the value provided is a value or percentage type
    """

    value: float
    type: ThresholdType


class OD_Base(ABC):
    """
    Base class for all outlier detection metrics in alibi-detect
    """

    @staticmethod
    def _update_kwargs_with_locals(kwargs_to_update, **kwargs):
        kwargs_to_update.update(
            {k: v for k, v in kwargs.items() if k != "self" and v is not None}
        )

    @abstractmethod
    def set_prediction_args(self) -> None:
        """Abstract method to set specific prediction arguments for the detector"""

    @property
    @abstractmethod
    def _default_predict_kwargs(self) -> dict:
        """Abstract method for the default prediction arguments for the detector"""

    def __init__(
        self,
        alibi_detect_class: type,
        model_param_name: str,
        flatten_dataset: bool,
        dataset_type: Optional[type] = None,
    ):
        """
        Constructor method

        Parameters
        ----------
        detector : Any, default None
            A model used for outlier detection after being trained on clean data
        alibi_detect_class : type
            This is the alibi_detect outlier detection class to instantiate
        flatten_dataset : bool, default False
            Flag to flatten a dataset to shape (B, H * W * C)
        dataset_type : Optional[type], default None
            Type used for formatting the dataset
        """
        self.detector: Any = None
        self.is_trained: bool = False

        self._alibi_detect_class = alibi_detect_class
        self._model_param_name = model_param_name
        self._flatten_dataset = flatten_dataset
        self._dataset_type = dataset_type

        self._model_kwargs = dict()
        self._predict_kwargs = self._default_predict_kwargs
        self._default_batch_size = 64

        self._input_shape: Tuple[int, int, int]

    # Train the alibi-detect metric on dataset
    def fit_dataset(
        self,
        images: np.ndarray,
        epochs: int = 3,
        threshold: Threshold = Threshold(95.0, ThresholdType.PERCENTAGE),
        batch_size: Optional[int] = None,
        verbose: bool = False,
    ) -> None:
        """
        Trains a model on a dataset containing that can be used
        for the detection of outliers in :meth:`evaluate`

        Parameters
        ----------
        images : np.ndarray
            A numpy array of images for the model to train on
        epochs : int, default 3
            Number of epochs to train the detector for
        threshold : Threshold, default Threshold(95.0, ThresholdType.PERCENTAGE)
            Sets the expected threshold of an outlier in the dataset
        batch_size : Optional[int], default None
            Batch size override to use during training
        verbose : bool, default False
            Flag to output logs from Alibi-Detect verbose mode

        Raises
        ------
        TypeError
            If the detector has not been initialized or loaded from path

        Note
        ----
        The supplied dataset should contain no outliers for maximum benefit
        """
        keras.backend.clear_session()
        if self.detector is None:
            self._input_shape = images[0].shape
            model = create_model(type(self).__name__, self._input_shape)
            self.detector = self._alibi_detect_class(
                threshold=0,
                **{self._model_param_name: model},
            )

        # Autoencoders only need images, so extract from dataset and format
        images = self._format_images(images)
        # Train the autoencoder using only the formatted images
        self.detector.fit(
            images,
            epochs=epochs,
            batch_size=batch_size or self._default_batch_size,
            verbose=verbose,
        )

        # if save_path: save_detector(self.detector, save_path)
        self.is_trained: bool = True

        if threshold.type is ThresholdType.PERCENTAGE:
            self.detector.infer_threshold(
                images,
                threshold_perc=threshold.value,
                batch_size=batch_size or self._default_batch_size,
            )
        else:
            self.detector.threshold = threshold.value

    def export_model(self, path: str):
        if self.detector is None:
            raise RuntimeError("The model must be initialized first")
        if self._model_param_name not in self.detector.__dict__.keys():
            raise ValueError("Member not found in detector")
        getattr(self.detector, self._model_param_name).save(path)

    def _check_dtype(self, images: np.ndarray):
        """
        Check if the dataset dtype fits with the required dtype of the model.
        None is used for any accepted type. Can extend to check for multiple types.

        Parameters
        ----------
        images : np.ndarray
            A numpy ndarray of images in the shape (B, H, W, C)
        dtype : Type
            The preferred dtype of the array

        Raises
        ------
        TypeError
            If the dataset is not of type np.ndarray

            If the dataset is not of the preferred dtype
        """

        if self._dataset_type is None:
            return

        # Use ndarray type conversion function
        if not isinstance(images, np.ndarray):
            raise TypeError("Dataset should be of type: np.ndarray")

        if not images.dtype.type == self._dataset_type:
            raise TypeError(
                f"Dataset values should be of type {self._dataset_type}, \
                not {images.dtype.type}"
            )

    def _check_image_shape(self, images: np.ndarray):
        """
        Verifies that a dataset has the same shape as a reference shape.
        Use this when fitting or evaluating a model on a dataset, to
        ensure that the dataset's dims are compatible with that which the model
        was built for.

        Parameters
        ----------
        image : np.ndarray
            An array of images in shape (B, H, W, C)

        Raises
        -------
        TypeError
            Raised if the dims of the last 3 channels of dataset
            don't line up with reference_input_shape
        """
        if self.detector is None:
            return
        if images.shape[-3:] != self._input_shape:
            raise ValueError(
                f"Model was initialized on dataset shape \
                (W,H,C)={self._input_shape}, \
                but provided dataset has shape (W,H,C)={images.shape[-3:]}"
            )

    def _format_images(self, images: np.ndarray) -> np.ndarray:
        """
        Formats images such that it fits the required datatype and shape

        Parameters
        ----------
        images : np.ndarray
            A numpy array of images in shape (B, H, W, C)

        Returns
        -------
        np.ndarray
            A numpy array of formatted images

        Note
        ----
        - Some metric libraries want a dataset in a particular format
            (e.g. float32, or flattened)
        - Override this to set the standard dataset formatting.
        """
        self._check_image_shape(images)
        self._check_dtype(images)

        if self._flatten_dataset:
            images = np.reshape(images, (len(images), math.prod(np.shape(images[0]))))
        return images

    def evaluate(
        self,
        images: np.ndarray,
    ) -> Dict[str, np.ndarray]:
        """
        Evaluate the outlier detector metric on a dataset.

        Parameters
        ----------
        images : np.ndarray
            The images to detect outliers on.

        Returns
        -------
        Dict[str, np.ndarray]
            is_outlier
            feature_score
            instance_score
        """
        if self.detector is None or not self.is_trained:
            raise TypeError(
                "Error: tried to evaluate a metric that is not trained. \
                    Try calling metric.fit_dataset(data)"
            )

        # Cast and flatten images
        images = self._format_images(images)

        self._predict_kwargs.update({"X": images})

        predictions = self.detector.predict(**self._predict_kwargs)

        output = {
            "is_outlier": predictions["data"]["is_outlier"].tolist(),
            "feature_score": predictions["data"]["feature_score"],
            "instance_score": predictions["data"]["instance_score"],
        }

        return output


class OD_AE(OD_Base):
    """
    Autoencoder-based outlier detector, using `alibi-detect ae. <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/ae.html>`_
    """  # noqa E501

    def __init__(self):
        super().__init__(
            alibi_detect_class=_alibi_detect.od.OutlierAE,
            model_param_name="ae",
            flatten_dataset=False,
            dataset_type=None,
        )

    def set_prediction_args(
        self,
        outlier_type: Optional[OutlierType] = None,
        outlier_perc: Optional[float] = None,
        return_feature_score: Optional[bool] = None,
        return_instance_score: Optional[bool] = None,
        batch_size: Optional[int] = None,
    ) -> None:
        """
        Sets additional arguments to be used during prediction.

        Note
        ----
        Visit `alibi-detect ae <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/ae.html#Detect>`_ for additional information on prediction parameters.
        """  # noqa E501
        self._update_kwargs_with_locals(self._predict_kwargs, **locals())

    @property
    def _default_predict_kwargs(self) -> dict:
        return {
            "outlier_type": OutlierType.INSTANCE,
            "outlier_perc": 75,
            "return_feature_score": True,
            "return_instance_score": True,
            "batch_size": 64,
        }


class OD_AEGMM(OD_Base):
    """
    Gaussian Mixture Model Autoencoder-based outlier detector,
    using alibi-detect aegmm. `<https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/aegmm.html>`_


    The model used by this class is :py:class:`daml.models.AEGMM`
    """  # noqa E501

    def __init__(self):
        super().__init__(
            alibi_detect_class=_alibi_detect.od.OutlierAEGMM,
            model_param_name="aegmm",
            flatten_dataset=True,
            dataset_type=np.float32,
        )

    def set_prediction_args(
        self,
        return_instance_score: Optional[bool] = None,
    ) -> None:
        """
        Sets additional arguments to be used during prediction.

        Note
        ----
        Visit `alibi-detect aegmm <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/aegmm.html#Detect>`_ for additional information on prediction parameters.
        """  # noqa E501
        self._update_kwargs_with_locals(self._predict_kwargs, **locals())

    @property
    def _default_predict_kwargs(self) -> dict:
        return {
            "return_instance_score": True,
            "batch_size": 64,
        }


class OD_LLR(OD_Base):
    """
    Log likelihood Ratio (LLR) outlier detector,
    using `alibi-detect llr. <https://docs.seldon.io/projects/alibi-detect/en/latest/examples/od_llr_mnist.html>`_


    The model used by this class is :py:class:`daml.models.LLR`
    """  # noqa E501

    def __init__(self):
        super().__init__(
            alibi_detect_class=_alibi_detect.od.LLR,
            model_param_name="model",
            flatten_dataset=False,
            dataset_type=np.float32,
        )

    def set_prediction_args(
        self,
        outlier_type: Optional[OutlierType] = None,
        return_instance_score: Optional[bool] = None,
    ) -> None:
        """
        Sets additional arguments to be used during prediction.

        Note
        ----
        Visit `alibi-detect llr <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/llr.html#Detect>`_ for additional information on prediction parameters.
        """  # noqa E501
        self._update_kwargs_with_locals(self._predict_kwargs, **locals())

    @property
    def _default_predict_kwargs(self) -> dict:
        return {
            "outlier_type": OutlierType.INSTANCE,
            "return_instance_score": True,
            "batch_size": 64,
        }


class OD_VAE(OD_Base):
    """
    Variational Autoencoder-based outlier detector,
    using `alibi-detect vae. <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/vae.html>`_

    The model used by this class is :py:class:`daml.models.VAE`
    """  # noqa E501

    def __init__(self):
        super().__init__(
            alibi_detect_class=_alibi_detect.od.OutlierVAE,
            model_param_name="vae",
            flatten_dataset=False,
            dataset_type=None,
        )

    def set_prediction_args(
        self,
        outlier_type: Optional[OutlierType] = None,
        outlier_perc: Optional[float] = None,
        return_feature_score: Optional[bool] = None,
        return_instance_score: Optional[bool] = None,
        batch_size: Optional[int] = None,
    ) -> None:
        """
        Sets additional arguments to be used during prediction.

        Note
        ----
        Visit `alibi-detect vae <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/vae.html#Detect>`_ for additional information on prediction parameters.
        """  # noqa E501
        self._update_kwargs_with_locals(self._predict_kwargs, **locals())

    @property
    def _default_predict_kwargs(self) -> dict:
        return {
            "outlier_type": OutlierType.INSTANCE,
            "outlier_perc": 75,
            "return_feature_score": True,
            "return_instance_score": True,
            "batch_size": 64,
        }


class OD_VAEGMM(OD_Base):
    """
    Variational Gaussian Mixture Model Autoencoder-based outlier detector,
    using `alibi-detect vaegmm. <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/vaegmm.html>`_


    The model used by this class is :py:class:`daml.models.VAEGMM`
    """  # noqa E501

    def __init__(self):
        super().__init__(
            alibi_detect_class=_alibi_detect.od.OutlierVAEGMM,
            model_param_name="vaegmm",
            flatten_dataset=True,
            dataset_type=np.float32,
        )

    def set_prediction_args(
        self,
        return_instance_score: Optional[bool] = None,
    ) -> None:
        """
        Sets additional arguments to be used during prediction.

        Note
        ----
        Visit `alibi-detect vaegmm <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/vaegmm.html#Detect>`_ for additional information on prediction parameters.
        """  # noqa E501
        self._update_kwargs_with_locals(self._predict_kwargs, **locals())

    @property
    def _default_predict_kwargs(self) -> dict:
        return {"return_instance_score": True, "batch_size": 64}



from typing import Optional

import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow.keras.layers import Flatten
from tensorflow.keras.losses import categorical_crossentropy, kld

from daml._alibi_detect.models.tensorflow.gmm import gmm_energy, gmm_params


def elbo(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    cov_full: Optional[tf.Tensor] = None,
    cov_diag: Optional[tf.Tensor] = None,
    sim: Optional[float] = None,
) -> tf.Tensor:
    """
    Compute ELBO loss. The covariance matrix can be specified by passing the full covariance matrix, the matrix
    diagonal, or a scale identity multiplier. Only one of these should be specified. If none are specified, the
    identity matrix is used.

    Parameters
    ----------
    y_true
        Labels.
    y_pred
        Predictions.
    cov_full
        Full covariance matrix.
    cov_diag
        Diagonal (variance) of covariance matrix.
    sim
        Scale identity multiplier.

    Returns
    -------
    ELBO loss value.

    Example
    -------
    >>> import tensorflow as tf
    >>> from alibi_detect.models.tensorflow.losses import elbo
    >>> y_true = tf.constant([[0.0, 1.0], [1.0, 0.0]])
    >>> y_pred = tf.constant([[0.1, 0.9], [0.8, 0.2]])
    >>> # Specifying scale identity multiplier
    >>> elbo(y_true, y_pred, sim=1.0)
    >>> # Specifying covariance matrix diagonal
    >>> elbo(y_true, y_pred, cov_diag=tf.ones(2))
    >>> # Specifying full covariance matrix
    >>> elbo(y_true, y_pred, cov_full=tf.eye(2))
    """
    if len([x for x in [cov_full, cov_diag, sim] if x is not None]) > 1:
        raise ValueError("Only one of cov_full, cov_diag or sim should be specified.")

    y_pred_flat = Flatten()(y_pred)
    if isinstance(cov_full, tf.Tensor):
        y_mn = tfp.distributions.MultivariateNormalFullCovariance(
            y_pred_flat, covariance_matrix=cov_full
        )
    else:
        if sim:
            cov_diag = sim * tf.ones(y_pred_flat.shape[-1])
        y_mn = tfp.distributions.MultivariateNormalDiag(
            y_pred_flat, scale_diag=cov_diag
        )
    loss = -tf.reduce_mean(y_mn.log_prob(Flatten()(y_true)))
    return loss


def loss_aegmm(
    x_true: tf.Tensor,
    x_pred: tf.Tensor,
    z: tf.Tensor,
    gamma: tf.Tensor,
    w_energy: float = 0.1,
    w_cov_diag: float = 0.005,
) -> tf.Tensor:
    """
    Loss function used for OutlierAEGMM.

    Parameters
    ----------
    x_true
        Batch of instances.
    x_pred
        Batch of reconstructed instances by the autoencoder.
    z
        Latent space values.
    gamma
        Membership prediction for mixture model components.
    w_energy
        Weight on sample energy loss term.
    w_cov_diag
        Weight on covariance regularizing loss term.

    Returns
    -------
    Loss value.
    """
    recon_loss = tf.reduce_mean((x_true - x_pred) ** 2)
    phi, mu, cov, L, log_det_cov = gmm_params(z, gamma)
    sample_energy, cov_diag = gmm_energy(
        z, phi, mu, cov, L, log_det_cov, return_mean=True
    )
    loss = recon_loss + w_energy * sample_energy + w_cov_diag * cov_diag
    return loss


def loss_vaegmm(
    x_true: tf.Tensor,
    x_pred: tf.Tensor,
    z: tf.Tensor,
    gamma: tf.Tensor,
    w_recon: float = 1e-7,
    w_energy: float = 0.1,
    w_cov_diag: float = 0.005,
    cov_full: tf.Tensor = None,
    cov_diag: tf.Tensor = None,
    sim: float = 0.05,
) -> tf.Tensor:
    """
    Loss function used for OutlierVAEGMM.

    Parameters
    ----------
    x_true
        Batch of instances.
    x_pred
        Batch of reconstructed instances by the variational autoencoder.
    z
        Latent space values.
    gamma
        Membership prediction for mixture model components.
    w_recon
        Weight on elbo loss term.
    w_energy
        Weight on sample energy loss term.
    w_cov_diag
        Weight on covariance regularizing loss term.
    cov_full
        Full covariance matrix.
    cov_diag
        Diagonal (variance) of covariance matrix.
    sim
        Scale identity multiplier.

    Returns
    -------
    Loss value.
    """
    recon_loss = elbo(x_true, x_pred, cov_full=cov_full, cov_diag=cov_diag, sim=sim)
    phi, mu, cov, L, log_det_cov = gmm_params(z, gamma)
    sample_energy, cov_diag = gmm_energy(z, phi, mu, cov, L, log_det_cov)
    loss = w_recon * recon_loss + w_energy * sample_energy + w_cov_diag * cov_diag
    return loss


def loss_adv_ae(
    x_true: tf.Tensor,
    x_pred: tf.Tensor,
    model: tf.keras.Model = None,
    model_hl: list = None,
    w_model: float = 1.0,
    w_recon: float = 0.0,
    w_model_hl: list = None,
    temperature: float = 1.0,
) -> tf.Tensor:
    """
    Loss function used for AdversarialAE.

    Parameters
    ----------
    x_true
        Batch of instances.
    x_pred
        Batch of reconstructed instances by the autoencoder.
    model
        A trained tf.keras model with frozen layers (layers.trainable = False).
    model_hl
        List with tf.keras models used to extract feature maps and make predictions on hidden layers.
    w_model
        Weight on model prediction loss term.
    w_recon
        Weight on MSE reconstruction error loss term.
    w_model_hl
        Weights assigned to the loss of each model in model_hl.
    temperature
        Temperature used for model prediction scaling.
        Temperature <1 sharpens the prediction probability distribution.

    Returns
    -------
    Loss value.
    """
    y_true = model(x_true)
    y_pred = model(x_pred)

    # apply temperature scaling
    if temperature != 1.0:
        y_true = y_true ** (1 / temperature)
        y_true = y_true / tf.reshape(tf.reduce_sum(y_true, axis=-1), (-1, 1))

    # compute K-L divergence loss
    loss_kld = kld(y_true, y_pred)
    std_kld = tf.math.reduce_std(loss_kld)
    loss = tf.reduce_mean(loss_kld)

    # add loss from optional K-L divergences extracted from hidden layers
    if isinstance(model_hl, list):
        if w_model_hl is None:
            w_model_hl = list(tf.ones(len(model_hl)))
        for m, w in zip(model_hl, w_model_hl):
            h_true = m(x_true)
            h_pred = m(x_pred)
            loss_kld_hl = tf.reduce_mean(kld(h_true, h_pred))
            loss += tf.constant(w) * loss_kld_hl
    loss *= w_model

    # add optional reconstruction loss
    if w_recon > 0.0:
        loss_recon = (x_true - x_pred) ** 2
        std_recon = tf.math.reduce_std(loss_recon)
        w_scale = std_kld / (std_recon + 1e-10)
        loss_recon = w_recon * w_scale * tf.reduce_mean(loss_recon)
        loss += loss_recon
        return loss
    else:
        return loss


def loss_distillation(
    x_true: tf.Tensor,
    y_pred: tf.Tensor,
    model: tf.keras.Model = None,
    loss_type: str = "kld",
    temperature: float = 1.0,
) -> tf.Tensor:
    """
    Loss function used for Model Distillation.

    Parameters
    ----------
    x_true
        Batch of data points.
    y_pred
        Batch of prediction from the distilled model.
    model
        tf.keras model.
    loss_type
        Type of loss for distillation. Supported 'kld', 'xent.
    temperature
        Temperature used for model prediction scaling.
        Temperature <1 sharpens the prediction probability distribution.

    Returns
    -------
    Loss value.
    """
    y_true = model(x_true)
    # apply temperature scaling
    if temperature != 1.0:
        y_true = y_true ** (1 / temperature)
        y_true = y_true / tf.reshape(tf.reduce_sum(y_true, axis=-1), (-1, 1))

    if loss_type == "kld":
        loss_dist = kld(y_true, y_pred)
    elif loss_type == "xent":
        loss_dist = categorical_crossentropy(y_true, y_pred, from_logits=False)
    else:
        raise NotImplementedError

    # compute K-L divergence loss
    loss = tf.reduce_mean(loss_dist)

    return loss



from typing import Any, List, Union

import torch
import torch.nn as nn
from torch.optim import Adam
from torch.utils.data import DataLoader, Dataset

torch.manual_seed(0)


def get_images_from_batch(batch: Any) -> Any:
    """Extracts images from a batch of collated data by DataLoader"""
    if isinstance(batch, (list, tuple)):
        imgs = batch[0]
    else:
        imgs = batch
    return imgs


class AETrainer:
    def __init__(
        self,
        model: nn.Module,
        device: Union[str, torch.device] = "auto",
        batch_size: int = 8,
    ):
        """
        model : nn.Module
            Model to be trained
        device : str | torch.device, default "cpu"
            Hardware device for model, optimizer, and data to run on
        batch_size : int, default 8
            Number of images to group together in `torch.utils.data.DataLoader`
        """
        if device == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device
        self.model = model.to(device)
        self.batch_size = batch_size

    def train(self, dataset: Dataset, epochs: int = 25) -> List[float]:
        """
        Basic training function for Autoencoder models for reconstruction tasks

        Uses `torch.optim.Adam` and `torch.nn.MSELoss` as default hyperparameters

        Parameters
        ----------
        dataset : Dataset
            Torch Dataset containing images in the first return position
        epochs : int, default 25
            Number of full training loops

        Note
        ----
        To replace this function with a custom function, do
            AETrainer.train = custom_function
        """
        # Setup training
        self.model.train()
        dataloader = DataLoader(dataset, batch_size=self.batch_size)
        opt = Adam(self.model.parameters(), lr=0.001)
        criterion = nn.MSELoss().to(self.device)
        # Record loss
        loss_history: List[float] = []
        batch_idx: int = 0

        for _ in range(epochs):
            epoch_loss: float = 0
            for batch_idx, batch in enumerate(dataloader):
                imgs = get_images_from_batch(batch)
                imgs = imgs.to(self.device)
                # Zero your gradients for every batch!
                opt.zero_grad()

                # Make predictions for this batch
                pred = self.model(imgs)

                # Compute the loss and its gradients
                loss = criterion(pred, imgs)
                loss.backward()

                # Adjust learning weights
                opt.step()

                # Gather data and report
                epoch_loss += loss.item()
            # Will take the average from all batches
            epoch_loss /= batch_idx + 1
            loss_history.append(epoch_loss)

        return loss_history

    @torch.no_grad
    def eval(self, dataset: Dataset) -> float:
        """
        Basic evaluation function for Autoencoder models for reconstruction tasks

        Uses `torch.optim.Adam` and `torch.nn.MSELoss` as default hyperparameters

        Parameters
        ----------
        dataset : Dataset
            Torch Dataset containing images in the first return position

        Returns
        -------
        float
            Total reconstruction loss over all data

        Note
        ----
        To replace this function with a custom function, do
            AETrainer.eval = custom_function
        """
        self.model.eval()
        dataloader = DataLoader(dataset, batch_size=self.batch_size)
        criterion = nn.MSELoss().to(self.device)
        total_loss: float = 0.0
        batch_idx: int = 0

        for batch_idx, batch in enumerate(dataloader):
            imgs = get_images_from_batch(batch)
            imgs = imgs.to(self.device)
            pred = self.model(imgs)
            loss = criterion(pred, imgs)
            total_loss += loss.item()
        return total_loss / (batch_idx + 1)

    @torch.no_grad
    def encode(self, dataset: Dataset) -> torch.Tensor:
        """
        Encode data through model if it has an encode attribute,
        otherwise passes data through model.forward

        Parameters
        ----------
        dataset: Dataset
            Dataset containing images to be encoded by the model

        Returns
        -------
        torch.Tensor
            Data encoded by the model
        """
        self.model.eval()
        dl = DataLoader(dataset, batch_size=self.batch_size)
        encodings = torch.Tensor([])

        # Get encode function if defined
        if getattr(self.model, "encode", None) is not None:
            encode_func = self.model.encode
        else:
            encode_func = self.model.forward
        # Accumulate encodings from batches
        for batch in dl:
            imgs = get_images_from_batch(batch)
            imgs = imgs.to(self.device)
            embeddings = encode_func(imgs).to("cpu")
            if len(encodings):
                encodings = torch.vstack((encodings, embeddings))
            else:
                encodings = embeddings

        return encodings


class AriaAutoencoder(nn.Module):
    def __init__(self, channels=3):
        super().__init__()
        self.encoder = Encoder(channels)
        self.decoder = Decoder(channels)

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x

    def encode(self, x):
        return self.encoder(x)


class Encoder(nn.Module):
    def __init__(self, channels=3):
        super().__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(channels, 256, 2, stride=1, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(256, 128, 2, stride=1, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(128, 64, 2, stride=1),
        )

    def forward(self, x):
        return self.encoder(x)


class Decoder(nn.Module):
    def __init__(self, channels):
        super().__init__()
        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(64, 128, 2, stride=1),
            nn.ReLU(),
            nn.ConvTranspose2d(128, 256, 2, stride=2),
            nn.ReLU(),
            nn.ConvTranspose2d(256, channels, 2, stride=2),
            nn.Sigmoid(),
        )

    def forward(self, x):
        return self.decoder(x)



from typing import Callable, Union

import numpy as np


def tokenize_transformer(
    x: Union[list, np.ndarray], tokenizer: Callable, max_len: int, backend: str
) -> dict:
    """
    Batch tokenizer for transformer models.

    Parameters
    ----------
    x
        Batch of instances.
    tokenizer
        Tokenizer for model.
    max_len
        Max token length.
    backend
        PyTorch ('pt') or TensorFlow ('tf') backend.

    Returns
    -------
    Tokenized instances.
    """
    return tokenizer(
        list(x),
        padding=True,
        truncation=True,
        max_length=max_len,
        return_tensors=backend,
    )



"""
This module contains the implementation of the
FR Test Statistic based estimate and the
KNN based estimate for the Bayes Error Rate

Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4)
https://arxiv.org/abs/1811.06419
"""

from typing import Callable, Dict, Literal, Tuple

import numpy as np
from scipy.sparse import coo_matrix

from daml._internal.metrics.base import EvaluateMixin, MethodsMixin

from .utils import compute_neighbors, get_classes_counts, minimum_spanning_tree


def _mst(X: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
    M, N = get_classes_counts(y)

    tree = coo_matrix(minimum_spanning_tree(X))
    matches = np.sum([y[tree.row[i]] != y[tree.col[i]] for i in range(N - 1)])
    deltas = matches / (2 * N)
    upper = 2 * deltas
    lower = ((M - 1) / (M)) * (1 - (1 - 2 * ((M) / (M - 1)) * deltas) ** 0.5)
    return upper, lower


def _knn(X: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
    M, N = get_classes_counts(y)

    # All features belong on second dimension
    X = X.reshape((X.shape[0], -1))
    nn_indices = compute_neighbors(X, X)
    deltas = float(np.count_nonzero(y[nn_indices] - y) / (2 * N))
    upper = 2 * deltas
    lower = ((M - 1) / (M)) * (1 - (1 - 2 * ((M) / (M - 1)) * deltas) ** 0.5)
    return upper, lower


_METHODS = Literal["MST", "KNN"]
_FUNCTION = Callable[[np.ndarray, np.ndarray], Tuple[float, float]]


class BER(EvaluateMixin, MethodsMixin[_METHODS, _FUNCTION]):
    """
    An estimator for Multi-class Bayes Error Rate using FR or KNN test statistic basis

    Parameters
    ----------
    data : np.ndarray
        Array of images or image embeddings
    labels : np.ndarray
        Array of labels for each image or image embedding
    method : Literal["MST", "KNN"], default "MST"
        Method to use when estimating the Bayes error rate

    See Also
    --------
    `Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4) <https://arxiv.org/abs/1811.06419>`_

    """  # noqa F401

    def __init__(
        self,
        data: np.ndarray,
        labels: np.ndarray,
        method: _METHODS = "MST",
    ) -> None:
        self.data = data
        self.labels = labels
        self._set_method(method)

    @classmethod
    def _methods(
        cls,
    ) -> Dict[str, _FUNCTION]:
        return {"MST": _mst, "KNN": _knn}

    def evaluate(self) -> Dict[str, float]:
        """
        Calculates the Bayes Error Rate estimate using the provided method

        Returns
        -------
        Dict[str, float]
            ber : float
                The estimated lower bounds of the Bayes Error Rate
            ber_lower : float
                The estimated upper bounds of the Bayes Error Rate

        Raises
        ------
        ValueError
            If unique classes M < 2
        """
        upper, lower = self._method(self.data, self.labels)
        return {"ber": upper, "ber_lower": lower}



import logging
from functools import partial
from typing import Callable, Dict, Tuple, Union

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model
from tensorflow_probability.python.distributions.distribution import Distribution

from daml._alibi_detect.base import (
    BaseDetector,
    FitMixin,
    ThresholdMixin,
    outlier_prediction_dict,
)
from daml._alibi_detect.models.tensorflow.pixelcnn import PixelCNN
from daml._alibi_detect.models.tensorflow.trainer import trainer
from daml._alibi_detect.utils._types import OptimizerTF
from daml._alibi_detect.utils.tensorflow.perturbation import mutate_categorical
from daml._alibi_detect.utils.tensorflow.prediction import predict_batch

logger = logging.getLogger(__name__)


def build_model(
    dist: Union[Distribution, PixelCNN], input_shape: tuple = None, filepath: str = None
) -> Tuple[tf.keras.Model, Union[Distribution, PixelCNN]]:
    """
    Create tf.keras.Model from TF distribution.

    Parameters
    ----------
    dist
        TensorFlow distribution.
    input_shape
        Input shape of the model.
    filepath
        File to load model weights from.

    Returns
    -------
    TensorFlow model.
    """
    x_in = Input(shape=input_shape)
    log_prob = dist.log_prob(x_in)
    model = Model(inputs=x_in, outputs=log_prob)
    model.add_loss(-tf.reduce_mean(log_prob))
    if isinstance(filepath, str):
        model.load_weights(filepath)
    return model, dist


class LLR(BaseDetector, FitMixin, ThresholdMixin):
    def __init__(
        self,
        threshold: float = None,
        model: Union[tf.keras.Model, Distribution, PixelCNN] = None,
        model_background: Union[tf.keras.Model, Distribution, PixelCNN] = None,
        log_prob: Callable = None,
        sequential: bool = False,
        data_type: str = None,
    ) -> None:
        """
        Likelihood Ratios for Out-of-Distribution Detection. Ren, J. et al. NeurIPS 2019.
        https://arxiv.org/abs/1906.02845

        Parameters
        ----------
        threshold
            Threshold used for the likelihood ratio (LLR) to determine outliers.
        model
            Generative model, defaults to PixelCNN.
        model_background
            Optional model for the background. Only needed if it is different from `model`.
        log_prob
            Function used to evaluate log probabilities under the model
            if the model does not have a `log_prob` function.
        sequential
            Whether the data is sequential. Used to create targets during training.
        data_type
            Optionally specify the data type (tabular, image or time-series). Added to metadata.
        """
        super().__init__()

        if threshold is None:
            logger.warning(
                "No threshold level set. Need to infer threshold using `infer_threshold`."
            )

        self.has_log_prob = True if hasattr(model, "log_prob") else False
        self.sequential = sequential
        self.log_prob = log_prob
        self.threshold = threshold

        # semantic model trained on original data
        self.dist_s = model
        # background model trained on perturbed data
        if model_background is None:
            try:
                self.dist_b = model.copy()
            except AttributeError:
                self.dist_b = tf.keras.models.clone_model(model)
        else:
            self.dist_b = model_background

        # set metadata
        self.meta["detector_type"] = "outlier"
        self.meta["data_type"] = data_type
        self.meta["online"] = False

    def fit(
        self,
        X: np.ndarray,
        mutate_fn: Callable = mutate_categorical,
        mutate_fn_kwargs: dict = {"rate": 0.2, "seed": 0, "feature_range": (0, 255)},
        mutate_batch_size: int = int(1e10),
        loss_fn: tf.keras.losses = None,
        loss_fn_kwargs: dict = None,
        optimizer: OptimizerTF = tf.keras.optimizers.Adam,
        epochs: int = 20,
        batch_size: int = 64,
        verbose: bool = True,
        log_metric: Tuple[str, "tf.keras.metrics"] = None,
        callbacks: tf.keras.callbacks = None,
    ) -> None:
        """
        Train semantic and background generative models.

        Parameters
        ----------
        X
            Training batch.
        mutate_fn
            Mutation function used to generate the background dataset.
        mutate_fn_kwargs
            Kwargs for the mutation function used to generate the background dataset.
            Default values set for an image dataset.
        mutate_batch_size
            Batch size used to generate the mutations for the background dataset.
        loss_fn
            Loss function used for training.
        loss_fn_kwargs
            Kwargs for loss function.
        optimizer
            Optimizer used for training.
        epochs
            Number of training epochs.
        batch_size
            Batch size used for training.
        verbose
            Whether to print training progress.
        log_metric
            Additional metrics whose progress will be displayed if verbose equals True.
        callbacks
            Callbacks used during training.
        """
        input_shape = X.shape[1:]
        optimizer = optimizer() if isinstance(optimizer, type) else optimizer
        # Separate into two separate optimizers, one for semantic model and one for background model
        optimizer_s = optimizer
        optimizer_b = optimizer.__class__.from_config(optimizer.get_config())

        # training arguments
        kwargs = {
            "epochs": epochs,
            "batch_size": batch_size,
            "verbose": verbose,
            "callbacks": callbacks,
        }

        # create background data
        mutate_fn = partial(mutate_fn, **mutate_fn_kwargs)
        X_back = predict_batch(
            X, mutate_fn, batch_size=mutate_batch_size, dtype=X.dtype
        )

        # prepare sequential data
        if self.sequential and not self.has_log_prob:
            y, y_back = X[:, 1:], X_back[:, 1:]  # type: ignore
            X, X_back = X[:, :-1], X_back[:, :-1]  # type: ignore
        else:
            y, y_back = None, None

        # check if model needs to be built
        use_build = (
            True
            if self.has_log_prob and not isinstance(self.dist_s, tf.keras.Model)
            else False
        )

        if use_build:
            # build and train semantic model
            self.model_s = build_model(self.dist_s, input_shape)[0]
            self.model_s.compile(optimizer=optimizer_s)
            self.model_s.fit(X, **kwargs)
            # build and train background model
            self.model_b = build_model(self.dist_b, input_shape)[0]
            self.model_b.compile(optimizer=optimizer_b)
            self.model_b.fit(X_back, **kwargs)
        else:
            # update training arguments
            kwargs.update({"loss_fn_kwargs": loss_fn_kwargs, "log_metric": log_metric})

            # train semantic model
            args = [self.dist_s, loss_fn, X]
            kwargs.update({"y_train": y, "optimizer": optimizer_s})
            trainer(*args, **kwargs)

            # train background model
            args = [self.dist_b, loss_fn, X_back]
            kwargs.update({"y_train": y_back, "optimizer": optimizer_b})
            trainer(*args, **kwargs)

    def infer_threshold(
        self,
        X: np.ndarray,
        outlier_type: str = "instance",
        threshold_perc: float = 95.0,
        batch_size: int = int(1e10),
    ) -> None:
        """
        Update LLR threshold by a value inferred from the percentage of instances
        considered to be outliers in a sample of the dataset.

        Parameters
        ----------
        X
            Batch of instances.
        outlier_type
            Predict outliers at the 'feature' or 'instance' level.
        threshold_perc
            Percentage of sorted feature level outlier scores used to predict instance level outlier.
        batch_size
            Batch size for the generative model evaluations.
        """
        # compute outlier scores
        fscore, iscore = self.score(X, batch_size=batch_size)
        if outlier_type == "feature":
            outlier_score = fscore
        elif outlier_type == "instance":
            outlier_score = iscore
        else:
            raise ValueError(
                "`outlier_score` needs to be either `feature` or `instance`."
            )

        # update threshold
        self.threshold = np.percentile(outlier_score, threshold_perc)

    def logp(
        self,
        dist,
        X: np.ndarray,
        return_per_feature: bool = False,
        batch_size: int = int(1e10),
    ) -> np.ndarray:
        """
        Compute log probability of a batch of instances under the generative model.

        Parameters
        ----------
        dist
            Distribution of the model.
        X
            Batch of instances.
        return_per_feature
            Return log probability per feature.
        batch_size
            Batch size for the generative model evaluations.

        Returns
        -------
        Log probabilities.
        """
        logp_fn = partial(dist.log_prob, return_per_feature=return_per_feature)
        # TODO: TBD: can this be any of the other types from predict_batch? i.e. tf.Tensor or tuple
        return predict_batch(X, logp_fn, batch_size=batch_size)  # type: ignore[return-value]

    def logp_alt(
        self,
        model: tf.keras.Model,
        X: np.ndarray,
        return_per_feature: bool = False,
        batch_size: int = int(1e10),
    ) -> np.ndarray:
        """
        Compute log probability of a batch of instances using the log_prob function
        defined by the user.

        Parameters
        ----------
        model
            Trained model.
        X
            Batch of instances.
        return_per_feature
            Return log probability per feature.
        batch_size
            Batch size for the generative model evaluations.

        Returns
        -------
        Log probabilities.
        """
        if self.sequential:
            y, X = X[:, 1:], X[:, :-1]
        else:
            y = X.copy()
        y_preds = predict_batch(X, model, batch_size=batch_size)
        logp = self.log_prob(y, y_preds).numpy()
        if return_per_feature:
            return logp
        else:
            axis = tuple(np.arange(len(logp.shape))[1:])
            return np.mean(logp, axis=axis)

    def llr(
        self, X: np.ndarray, return_per_feature: bool, batch_size: int = int(1e10)
    ) -> np.ndarray:
        """
        Compute likelihood ratios.

        Parameters
        ----------
        X
            Batch of instances.
        return_per_feature
            Return likelihood ratio per feature.
        batch_size
            Batch size for the generative model evaluations.

        Returns
        -------
        Likelihood ratios.
        """
        logp_fn = self.logp if not isinstance(self.log_prob, Callable) else self.logp_alt  # type: ignore
        logp_s = logp_fn(
            self.dist_s, X, return_per_feature=return_per_feature, batch_size=batch_size
        )
        logp_b = logp_fn(
            self.dist_b, X, return_per_feature=return_per_feature, batch_size=batch_size
        )
        return logp_s - logp_b

    def feature_score(self, X: np.ndarray, batch_size: int = int(1e10)) -> np.ndarray:
        """Feature-level negative likelihood ratios."""
        return -self.llr(X, True, batch_size=batch_size)

    def instance_score(self, X: np.ndarray, batch_size: int = int(1e10)) -> np.ndarray:
        """Instance-level negative likelihood ratios."""
        return -self.llr(X, False, batch_size=batch_size)

    def score(
        self, X: np.ndarray, batch_size: int = int(1e10)
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Feature-level and instance-level outlier scores.
        The scores are equal to the negative likelihood ratios.
        """
        fscore = self.feature_score(X, batch_size=batch_size)
        iscore = self.instance_score(X, batch_size=batch_size)
        return fscore, iscore

    def predict(
        self,
        X: np.ndarray,
        outlier_type: str = "instance",
        batch_size: int = int(1e10),
        return_feature_score: bool = True,
        return_instance_score: bool = True,
    ) -> Dict[Dict[str, str], Dict[np.ndarray, np.ndarray]]:
        """
        Predict whether instances are outliers or not.

        Parameters
        ----------
        X
            Batch of instances.
        outlier_type
            Predict outliers at the 'feature' or 'instance' level.
        batch_size
            Batch size used when making predictions with the generative model.
        return_feature_score
            Whether to return feature level outlier scores.
        return_instance_score
            Whether to return instance level outlier scores.

        Returns
        -------
        Dictionary containing ``'meta'`` and ``'data'`` dictionaries.
            - ``'meta'`` has the model's metadata.
            - ``'data'`` contains the outlier predictions and both feature and instance level outlier scores.
        """
        # compute outlier scores
        fscore, iscore = self.score(X, batch_size=batch_size)
        if outlier_type == "feature":
            outlier_score = fscore
        elif outlier_type == "instance":
            outlier_score = iscore
        else:
            raise ValueError(
                "`outlier_score` needs to be either `feature` or `instance`."
            )

        # values above threshold are outliers
        outlier_pred = (outlier_score > self.threshold).astype(int)

        # populate output dict
        od = outlier_prediction_dict()
        od["meta"] = self.meta
        od["data"]["is_outlier"] = outlier_pred
        if return_feature_score:
            od["data"]["feature_score"] = fscore
        if return_instance_score:
            od["data"]["instance_score"] = iscore
        return od



from daml._internal.metrics.od import (
    OD_AE,
    OD_AEGMM,
    OD_LLR,
    OD_VAE,
    OD_VAEGMM,
    OutlierType,
    Threshold,
    ThresholdType,
)

__all__ = [
    "OD_AE",
    "OD_AEGMM",
    "OD_LLR",
    "OD_VAE",
    "OD_VAEGMM",
    "OutlierType",
    "Threshold",
    "ThresholdType",
]



from functools import partial
from typing import Callable, Tuple

import numpy as np
import tensorflow as tf


def trainer(
    model: tf.keras.Model,
    loss_fn: tf.keras.losses,
    x_train: np.ndarray,
    y_train: np.ndarray = None,
    dataset: tf.keras.utils.Sequence = None,
    optimizer: tf.keras.optimizers = tf.keras.optimizers.Adam,
    loss_fn_kwargs: dict = None,
    preprocess_fn: Callable = None,
    epochs: int = 20,
    reg_loss_fn: Callable = (lambda model: 0),
    batch_size: int = 64,
    buffer_size: int = 1024,
    verbose: bool = True,
    log_metric: Tuple[str, "tf.keras.metrics"] = None,
    callbacks: tf.keras.callbacks = None,
) -> None:
    """
    Train TensorFlow model.

    Parameters
    ----------
    model
        Model to train.
    loss_fn
        Loss function used for training.
    x_train
        Training data.
    y_train
        Training labels.
    dataset
        Training dataset which returns (x, y).
    optimizer
        Optimizer used for training.
    loss_fn_kwargs
        Kwargs for loss function.
    preprocess_fn
        Preprocessing function applied to each training batch.
    epochs
        Number of training epochs.
    reg_loss_fn
        Allows an additional regularisation term to be defined as reg_loss_fn(model)
    batch_size
        Batch size used for training.
    buffer_size
        Maximum number of elements that will be buffered when prefetching.
    verbose
        Whether to print training progress.
    log_metric
        Additional metrics whose progress will be displayed if verbose equals True.
    callbacks
        Callbacks used during training.
    """
    optimizer = optimizer() if isinstance(optimizer, type) else optimizer
    return_xy = (
        False
        if not isinstance(dataset, tf.keras.utils.Sequence) and y_train is None
        else True
    )
    if not isinstance(dataset, tf.keras.utils.Sequence):  # create dataset
        train_data = x_train if y_train is None else (x_train, y_train)
        dataset = tf.data.Dataset.from_tensor_slices(train_data)
        dataset = dataset.shuffle(buffer_size=buffer_size).batch(batch_size)
    n_minibatch = len(dataset)

    if loss_fn_kwargs:
        loss_fn = partial(loss_fn, **loss_fn_kwargs)

    # iterate over epochs
    for epoch in range(epochs):
        if verbose:
            pbar = tf.keras.utils.Progbar(n_minibatch, 1)
        if hasattr(dataset, "on_epoch_end"):
            dataset.on_epoch_end()
        loss_val_ma = 0.0
        for step, data in enumerate(dataset):
            x, y = data if return_xy else (data, None)
            if isinstance(preprocess_fn, Callable):  # type: ignore
                x = preprocess_fn(x)
            with tf.GradientTape() as tape:
                y_hat = model(x)
                y = x if y is None else y
                if isinstance(loss_fn, Callable):  # type: ignore
                    args = [y, y_hat] if tf.is_tensor(y_hat) else [y] + list(y_hat)
                    loss = loss_fn(*args)
                else:
                    loss = 0.0
                if model.losses:  # additional model losses
                    loss += sum(model.losses)
                loss += reg_loss_fn(model)  # alternative way they might be specified

            grads = tape.gradient(loss, model.trainable_weights)
            optimizer.apply_gradients(zip(grads, model.trainable_weights))
            if verbose:
                loss_val = loss.numpy()
                if loss_val.shape:
                    if loss_val.shape[0] != batch_size:
                        if len(loss_val.shape) == 1:
                            shape = (batch_size - loss_val.shape[0],)
                        elif len(loss_val.shape) == 2:
                            shape = (batch_size - loss_val.shape[0], loss_val.shape[1])  # type: ignore
                        add_mean = np.ones(shape) * loss_val.mean()
                        loss_val = np.r_[loss_val, add_mean]
                loss_val_ma = loss_val_ma + (loss_val - loss_val_ma) / (step + 1)
                pbar_values = [("loss_ma", loss_val_ma)]
                if log_metric is not None:
                    log_metric[1](y, y_hat)
                    pbar_values.append((log_metric[0], log_metric[1].result().numpy()))
                pbar.add(1, values=pbar_values)



import torch.nn as nn


class Conv(nn.Module):
    """
    Wrapper for conv modules, so we don't have to specify everything every time
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        k=1,
        s=1,
        p=0,
        activation="relu",
        norm="instance",
    ):
        super().__init__()
        conv = nn.Conv2d(in_channels, out_channels, kernel_size=k, stride=s, padding=p)
        norm = self.get_norm_func(norm=norm, out_channels=out_channels)
        act = self.get_activation_func(activation=activation)
        self.module = nn.Sequential(conv, norm, act)

    def get_norm_func(self, norm: str, out_channels) -> nn.Module:
        if norm == "batch":
            return nn.BatchNorm2d(out_channels)
        if norm == "instance":
            return nn.InstanceNorm2d(out_channels)
        if norm == "layer":
            return nn.LayerNorm(out_channels)
        return nn.Identity()

    def get_activation_func(self, activation: str) -> nn.Module:
        if activation == "selu":
            return nn.SELU()
        if activation == "relu":
            return nn.ReLU()
        if activation == "leaky":
            return nn.LeakyReLU()
        if activation == "tanh":
            return nn.Tanh()
        return nn.Identity()

    def forward(self, x):
        return self.module(x)



import logging
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf

logger = logging.getLogger(__name__)


def squared_pairwise_distance(
    x: tf.Tensor, y: tf.Tensor, a_min: float = 1e-30, a_max: float = 1e30
) -> tf.Tensor:
    """
    TensorFlow pairwise squared Euclidean distance between samples x and y.

    Parameters
    ----------
    x
        Batch of instances of shape [Nx, features].
    y
        Batch of instances of shape [Ny, features].
    a_min
        Lower bound to clip distance values.
    a_max
        Upper bound to clip distance values.

    Returns
    -------
    Pairwise squared Euclidean distance [Nx, Ny].
    """
    x2 = tf.reduce_sum(x**2, axis=-1, keepdims=True)
    y2 = tf.reduce_sum(y**2, axis=-1, keepdims=True)
    dist = x2 + tf.transpose(y2, (1, 0)) - 2.0 * x @ tf.transpose(y, (1, 0))
    return tf.clip_by_value(dist, a_min, a_max)


def batch_compute_kernel_matrix(
    x: Union[list, np.ndarray, tf.Tensor],
    y: Union[list, np.ndarray, tf.Tensor],
    kernel: Union[Callable, tf.keras.Model],
    batch_size: int = int(1e10),
    preprocess_fn: Callable = None,
) -> tf.Tensor:
    """
    Compute the kernel matrix between x and y by filling in blocks of size
    batch_size x batch_size at a time.

    Parameters
    ----------
    x
        Reference set.
    y
        Test set.
    kernel
        tf.keras model
    batch_size
        Batch size used during prediction.
    preprocess_fn
        Optional preprocessing function for each batch.

    Returns
    -------
    Kernel matrix in the form of a tensorflow tensor
    """
    if type(x) != type(y):  # noqa: E721
        raise ValueError("x and y should be of the same type")

    n_x, n_y = len(x), len(y)
    n_batch_x, n_batch_y = int(np.ceil(n_x / batch_size)), int(
        np.ceil(n_y / batch_size)
    )

    k_is = []
    for i in range(n_batch_x):
        istart, istop = i * batch_size, min((i + 1) * batch_size, n_x)
        x_batch = x[istart:istop]
        if isinstance(preprocess_fn, Callable):  # type: ignore
            x_batch = preprocess_fn(x_batch)
        k_ijs = []
        for j in range(n_batch_y):
            jstart, jstop = j * batch_size, min((j + 1) * batch_size, n_y)
            y_batch = y[jstart:jstop]
            if isinstance(preprocess_fn, Callable):  # type: ignore
                y_batch = preprocess_fn(y_batch)
            k_ijs.append(kernel(x_batch, y_batch))
        k_is.append(tf.concat(k_ijs, axis=1))
    k_mat = tf.concat(k_is, axis=0)
    return k_mat


def mmd2_from_kernel_matrix(
    kernel_mat: tf.Tensor, m: int, permute: bool = False, zero_diag: bool = True
) -> tf.Tensor:
    """
    Compute maximum mean discrepancy (MMD^2) between 2 samples x and y from the
    full kernel matrix between the samples.

    Parameters
    ----------
    kernel_mat
        Kernel matrix between samples x and y.
    m
        Number of instances in y.
    permute
        Whether to permute the row indices. Used for permutation tests.
    zero_diag
        Whether to zero out the diagonal of the kernel matrix.

    Returns
    -------
    MMD^2 between the samples from the kernel matrix.
    """
    n = kernel_mat.shape[0] - m
    if zero_diag:
        kernel_mat = kernel_mat - tf.linalg.diag(tf.linalg.diag_part(kernel_mat))
    if permute:
        idx = np.random.permutation(kernel_mat.shape[0])
        kernel_mat = tf.gather(
            tf.gather(kernel_mat, indices=idx, axis=0), indices=idx, axis=1
        )
    k_xx, k_yy, k_xy = kernel_mat[:-m, :-m], kernel_mat[-m:, -m:], kernel_mat[-m:, :-m]
    c_xx, c_yy = 1 / (n * (n - 1)), 1 / (m * (m - 1))
    mmd2 = (
        c_xx * tf.reduce_sum(k_xx)
        + c_yy * tf.reduce_sum(k_yy)
        - 2.0 * tf.reduce_mean(k_xy)
    )
    return mmd2


def mmd2(x: tf.Tensor, y: tf.Tensor, kernel: Callable) -> float:
    """
    Compute MMD^2 between 2 samples.

    Parameters
    ----------
    x
        Batch of instances of shape [Nx, features].
    y
        Batch of instances of shape [Ny, features].
    kernel
        Kernel function.

    Returns
    -------
    MMD^2 between the samples x and y.
    """
    n, m = x.shape[0], y.shape[0]
    c_xx, c_yy = 1 / (n * (n - 1)), 1 / (m * (m - 1))
    k_xx, k_yy, k_xy = kernel(x, x), kernel(y, y), kernel(x, y)
    return (
        c_xx * (tf.reduce_sum(k_xx) - tf.linalg.trace(k_xx))
        + c_yy * (tf.reduce_sum(k_yy) - tf.linalg.trace(k_yy))
        - 2.0 * tf.reduce_mean(k_xy)
    )


def relative_euclidean_distance(
    x: tf.Tensor, y: tf.Tensor, eps: float = 1e-12, axis: int = -1
) -> tf.Tensor:
    """
    Relative Euclidean distance.

    Parameters
    ----------
    x
        Tensor used in distance computation.
    y
        Tensor used in distance computation.
    eps
        Epsilon added to denominator for numerical stability.
    axis
        Axis used to compute distance.

    Returns
    -------
    Tensor with relative Euclidean distance across specified axis.
    """
    denom = tf.concat(
        [
            tf.reshape(tf.norm(x, ord=2, axis=axis), (-1, 1)),
            tf.reshape(tf.norm(y, ord=2, axis=axis), (-1, 1)),
        ],
        axis=1,
    )
    dist = tf.norm(x - y, ord=2, axis=axis) / (tf.reduce_min(denom, axis=axis) + eps)
    return dist


def permed_lsdds(
    k_all_c: tf.Tensor,
    x_perms: List[tf.Tensor],
    y_perms: List[tf.Tensor],
    H: tf.Tensor,
    H_lam_inv: Optional[tf.Tensor] = None,
    lam_rd_max: float = 0.2,
    return_unpermed: bool = False,
) -> Union[Tuple[tf.Tensor, tf.Tensor], Tuple[tf.Tensor, tf.Tensor, tf.Tensor]]:
    """
    Compute LSDD estimates from kernel matrix across various ref and test window samples

    Parameters
    ----------
    k_all_c
        Kernel matrix of similarities between all samples and the kernel centers.
    x_perms
        List of B reference window index vectors
    y_perms
        List of B test window index vectors
    H
        Special (scaled) kernel matrix of similarities between kernel centers
    H_lam_inv
        Function of H corresponding to a particular regulariation parameter lambda.
    lam_rd_max
        The maximum relative difference between two estimates of LSDD that the regularization parameter
        lambda is allowed to cause. Defaults to 0.2. Only relevant if H_lam_inv is not supplied.
    return_unpermed
        Whether or not to return value corresponding to unpermed order defined by k_all_c

    Returns
    -------
    Vector of B LSDD estimates for each permutation, H_lam_inv which may have been inferred, and optionally \
    the unpermed LSDD estimate.
    """

    # Compute (for each bootstrap) the average distance to each kernel center (Eqn 7)
    k_xc_perms = tf.stack([tf.gather(k_all_c, x_inds) for x_inds in x_perms], axis=0)
    k_yc_perms = tf.stack([tf.gather(k_all_c, y_inds) for y_inds in y_perms], axis=0)
    h_perms = tf.reduce_mean(k_xc_perms, axis=1) - tf.reduce_mean(k_yc_perms, axis=1)

    if H_lam_inv is None:
        # We perform the initialisation for multiple candidate lambda values and pick the largest
        # one for which the relative difference (RD) between two difference estimates is below lambda_rd_max.
        # See Appendix A
        candidate_lambdas = [
            1 / (4**i) for i in range(10)
        ]  # TODO: More principled selection
        H_plus_lams = tf.stack(
            [
                H + tf.eye(H.shape[0], dtype=H.dtype) * can_lam
                for can_lam in candidate_lambdas
            ],
            axis=0,
        )
        H_plus_lam_invs = tf.transpose(
            tf.linalg.inv(H_plus_lams), [1, 2, 0]
        )  # lambdas last
        omegas = tf.einsum("jkl,bk->bjl", H_plus_lam_invs, h_perms)  # (Eqn 8)
        h_omegas = tf.einsum("bj,bjl->bl", h_perms, omegas)
        omega_H_omegas = tf.einsum(
            "bkl,bkl->bl", tf.einsum("bjl,jk->bkl", omegas, H), omegas
        )
        rds = tf.reduce_mean(1 - (omega_H_omegas / h_omegas), axis=0)
        less_than_rd_inds = tf.where(rds < lam_rd_max)
        if len(less_than_rd_inds) == 0:
            repeats = k_all_c.shape[0] - np.unique(k_all_c, axis=0).shape[0]
            if repeats > 0:
                msg = "Too many repeat instances for LSDD-based detection. \
                Try using MMD-based detection instead"
            else:
                msg = "Unknown error. Try using MMD-based detection instead"
            raise ValueError(msg)
        lambda_index = int(less_than_rd_inds[0])
        lam = candidate_lambdas[lambda_index]
        logger.info(
            f"Using lambda value of {lam:.2g} with RD of {float(rds[lambda_index]):.2g}"
        )
        H_plus_lam_inv = tf.linalg.inv(H + lam * tf.eye(H.shape[0], dtype=H.dtype))
        H_lam_inv = 2 * H_plus_lam_inv - (
            tf.transpose(H_plus_lam_inv, [1, 0]) @ H @ H_plus_lam_inv
        )  # (blw Eqn 11)

    # Now to compute an LSDD estimate for each permutation
    lsdd_perms = tf.reduce_sum(
        h_perms * tf.transpose(H_lam_inv @ tf.transpose(h_perms, [1, 0]), [1, 0]),
        axis=1,
    )  # (Eqn 11)

    if return_unpermed:
        n_x = x_perms[0].shape[0]
        h = tf.reduce_mean(k_all_c[:n_x], axis=0) - tf.reduce_mean(
            k_all_c[n_x:], axis=0
        )
        lsdd_unpermed = tf.reduce_sum(
            h[None, :] * tf.transpose(H_lam_inv @ h[:, None], [1, 0])
        )
        return lsdd_perms, H_lam_inv, lsdd_unpermed
    else:
        return lsdd_perms, H_lam_inv



from importlib.util import find_spec

from . import metrics

__version__ = "0.44.3"

__all__ = ["metrics"]

if find_spec("torch") is not None:  # pragma: no cover
    from . import models

    __all__ += ["models"]

del find_spec



from typing import Tuple

import numpy as np
import tensorflow as tf


def gmm_params(
    z: tf.Tensor, gamma: tf.Tensor
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    """
    Compute parameters of Gaussian Mixture Model.

    Parameters
    ----------
    z
        Observations.
    gamma
        Mixture probabilities to derive mixture distribution weights from.

    Returns
    -------
    phi
        Mixture component distribution weights.
    mu
        Mixture means.
    cov
        Mixture covariance.
    L
        Cholesky decomposition of `cov`.
    log_det_cov
        Log of the determinant of `cov`.
    """
    # compute gmm parameters phi, mu and cov
    N = gamma.shape[0]  # nb of samples in batch
    sum_gamma = tf.reduce_sum(gamma, 0)  # K
    phi = sum_gamma / N  # K
    mu = tf.reduce_sum(
        tf.expand_dims(gamma, -1) * tf.expand_dims(z, 1), 0
    ) / tf.expand_dims(
        sum_gamma, -1
    )  # K x D (D = latent_dim)
    z_mu = tf.expand_dims(z, 1) - tf.expand_dims(mu, 0)  # N x K x D
    z_mu_outer = tf.expand_dims(z_mu, -1) * tf.expand_dims(z_mu, -2)  # N x K x D x D
    cov = tf.reduce_sum(
        tf.expand_dims(tf.expand_dims(gamma, -1), -1) * z_mu_outer, 0
    ) / tf.expand_dims(
        tf.expand_dims(sum_gamma, -1), -1
    )  # K x D x D

    # cholesky decomposition of covariance and determinant derivation
    D = tf.shape(cov)[1]
    eps = 1e-6
    L = tf.linalg.cholesky(cov + tf.eye(D) * eps)  # K x D x D
    log_det_cov = 2.0 * tf.reduce_sum(tf.math.log(tf.linalg.diag_part(L)), 1)  # K

    return phi, mu, cov, L, log_det_cov


def gmm_energy(
    z: tf.Tensor,
    phi: tf.Tensor,
    mu: tf.Tensor,
    cov: tf.Tensor,
    L: tf.Tensor,
    log_det_cov: tf.Tensor,
    return_mean: bool = True,
) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    Compute sample energy from Gaussian Mixture Model.

    Parameters
    ----------
    z
        Observations.
    phi
        Mixture component distribution weights.
    mu
        Mixture means.
    cov
        Mixture covariance.
    L
        Cholesky decomposition of `cov`.
    log_det_cov
        Log of the determinant of `cov`.
    return_mean
        Take mean across all sample energies in a batch.

    Returns
    -------
    sample_energy
        The sample energy of the GMM.
    cov_diag
        The inverse sum of the diagonal components of the covariance matrix.
    """
    D = tf.shape(cov)[1]
    z_mu = tf.expand_dims(z, 1) - tf.expand_dims(mu, 0)  # N x K x D
    z_mu_T = tf.transpose(z_mu, perm=[1, 2, 0])  # K x D x N
    v = tf.linalg.triangular_solve(L, z_mu_T, lower=True)  # K x D x D

    # rewrite sample energy in logsumexp format for numerical stability
    logits = tf.math.log(tf.expand_dims(phi, -1)) - 0.5 * (
        tf.reduce_sum(tf.square(v), 1)
        + tf.cast(D, tf.float32) * tf.math.log(2.0 * np.pi)
        + tf.expand_dims(log_det_cov, -1)
    )  # K x N
    sample_energy = -tf.reduce_logsumexp(logits, axis=0)  # N

    if return_mean:
        sample_energy = tf.reduce_mean(sample_energy)

    # inverse sum of variances
    cov_diag = tf.reduce_sum(tf.divide(1, tf.linalg.diag_part(cov)))

    return sample_energy, cov_diag



import copy
import json
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Union

import numpy as np
from typing_extensions import Protocol, runtime_checkable

from daml._alibi_detect.version import __version__

DEFAULT_META: Dict = {
    "name": None,
    "online": None,  # true or false
    "data_type": None,  # tabular, image or time-series
    "version": None,
    "detector_type": None,  # drift, outlier or adversarial
}


def outlier_prediction_dict():
    data = {"instance_score": None, "feature_score": None, "is_outlier": None}
    return copy.deepcopy({"data": data, "meta": DEFAULT_META})


def adversarial_prediction_dict():
    data = {"instance_score": None, "is_adversarial": None}
    return copy.deepcopy({"data": data, "meta": DEFAULT_META})


def adversarial_correction_dict():
    data = {
        "instance_score": None,
        "is_adversarial": None,
        "corrected": None,
        "no_defense": None,
        "defense": None,
    }
    return copy.deepcopy({"data": data, "meta": DEFAULT_META})


def concept_drift_dict():
    data = {"is_drift": None, "distance": None, "p_val": None, "threshold": None}
    return copy.deepcopy({"data": data, "meta": DEFAULT_META})


class BaseDetector(ABC):
    """Base class for outlier, adversarial and drift detection algorithms."""

    def __init__(self):
        self.meta = copy.deepcopy(DEFAULT_META)
        self.meta["name"] = self.__class__.__name__
        self.meta["version"] = __version__

    def __repr__(self):
        return self.__class__.__name__

    @property
    def meta(self) -> Dict:
        return self._meta

    @meta.setter
    def meta(self, value: Dict):
        if not isinstance(value, dict):
            raise TypeError("meta must be a dictionary")
        self._meta = value

    @abstractmethod
    def score(self, X: np.ndarray):
        pass

    @abstractmethod
    def predict(self, X: np.ndarray):
        pass


class FitMixin(ABC):
    @abstractmethod
    def fit(self, *args, **kwargs) -> None: ...


class ThresholdMixin(ABC):
    @abstractmethod
    def infer_threshold(self, *args, **kwargs) -> None: ...


# "Large artefacts" - to save memory these are skipped in _set_config(), but added back in get_config()
# Note: The current implementation assumes the artefact is stored as a class attribute, and as a config field under
# the same name. Refactoring will be required if this assumption is to be broken.
LARGE_ARTEFACTS = ["x_ref", "c_ref", "preprocess_fn"]


class DriftConfigMixin:
    """
    A mixin class containing methods related to a drift detector's configuration dictionary.
    """

    config: Optional[dict] = None

    def get_config(
        self,
    ) -> (
        dict
    ):  # TODO - move to BaseDetector once config save/load implemented for non-drift
        """
        Get the detector's configuration dictionary.

        Returns
        -------
        The detector's configuration dictionary.
        """
        if self.config is not None:
            # Get config (stored in top-level self)
            cfg = self.config
            # Add large artefacts back to config
            for key in LARGE_ARTEFACTS:
                if key in cfg and hasattr(self._nested_detector, key):
                    cfg[key] = getattr(self._nested_detector, key)
            # Set x_ref_preprocessed flag
            # If no preprocess_at_init, always true!
            preprocess_at_init = getattr(
                self._nested_detector, "preprocess_at_init", True
            )
            cfg["x_ref_preprocessed"] = (
                preprocess_at_init and self._nested_detector.preprocess_fn is not None
            )
            return cfg
        else:
            raise NotImplementedError(
                "Getting a config (or saving via a config file) is not yet implemented for this"
                "detector"
            )

    @classmethod
    def from_config(cls, config: dict):
        """
        Instantiate a drift detector from a fully resolved (and validated) config dictionary.

        Parameters
        ----------
        config
            A config dictionary matching the schema's in :class:`~_alibi_detect.saving.schemas`.
        """
        # Check for existing version_warning. meta is pop'd as don't want to pass as arg/kwarg
        meta = config.pop("meta", None)
        meta = (
            {} if meta is None else meta
        )  # Needed because pydantic sets meta=None if it is missing from the config
        version_warning = meta.pop("version_warning", False)
        # Init detector
        detector = cls(**config)
        # Add version_warning
        detector.meta["version_warning"] = version_warning  # type: ignore[attr-defined]
        detector.config["meta"]["version_warning"] = version_warning
        return detector

    def _set_config(
        self, inputs: dict
    ):  # TODO - move to BaseDetector once config save/load implemented for non-drift
        """
        Set a detectors `config` attribute upon detector instantiation.

        Large artefacts are overwritten with `None` in order to avoid memory duplication. They're added back into
        the config later on by `get_config()`.

        Parameters
        ----------
        inputs
            The inputs (args/kwargs) given to the detector at instantiation.
        """
        # Set config metadata
        name = self.__class__.__name__

        # Init config dict
        self.config = {
            "name": name,
            "meta": {
                "version": __version__,
            },
        }

        # args and kwargs
        pop_inputs = ["self", "__class__", "__len__", "name", "meta"]
        [inputs.pop(k, None) for k in pop_inputs]

        # Overwrite any large artefacts with None to save memory. They'll be added back by get_config()
        for key in LARGE_ARTEFACTS:
            if key in inputs and hasattr(self._nested_detector, key):
                inputs[key] = None

        self.config.update(inputs)

    @property
    def _nested_detector(self):
        """
        The low-level nested detector.
        """
        detector = self._detector if hasattr(self, "_detector") else self
        detector = detector._detector if hasattr(detector, "_detector") else detector
        return detector


@runtime_checkable
class Detector(Protocol):
    """Type Protocol for all detectors.

    Used for typing legacy save and load functionality in `_alibi_detect.saving._tensorflow.saving.py`.

    Note
    ----
        This exists to distinguish between detectors with and without support for config saving and loading. Once all
        detector support this then this protocol will be removed.
    """

    meta: Dict

    def predict(self) -> Any: ...


@runtime_checkable
class ConfigurableDetector(Detector, Protocol):
    """Type Protocol for detectors that have support for saving via config.

    Used for typing save and load functionality in `_alibi_detect.saving.saving`.
    """

    def get_config(self) -> dict: ...

    @classmethod
    def from_config(cls, config: dict): ...

    def _set_config(self, inputs: dict): ...


@runtime_checkable
class StatefulDetectorOnline(ConfigurableDetector, Protocol):
    """Type Protocol for detectors that have support for save/loading of online state.

    Used for typing save and load functionality in `_alibi_detect.saving.saving`.
    """

    t: int = 0

    def save_state(self, filepath: Union[str, os.PathLike]): ...

    def load_state(self, filepath: Union[str, os.PathLike]): ...


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)
